import { readFileSync, writeFileSync } from 'fs'
import { resolve } from 'path'
import { cwd } from 'process'
import config from '@/config'
import type { FastifyInstance } from 'fastify'
import type { GenesisInfo } from './index-blocks'

// this file indexes data from json specs generated by a command like:
// duniter build-spec --chain gdev_dev 1> ./specs.json

// define gensis interfaces
// pallet / genesis config
export interface Genesis {
  system: any
  account: Accounts
  babe: any
  parameters: any
  balances: any
  authorityMembers: InitialAuthorities
  session: any
  grandpa: any
  imOnline: any
  authorityDiscovery: any
  sudo: any
  technicalCommittee: any
  universalDividend: any
  identity: Identities
  membership: Memberships
  cert: Certs
  smithMembership: Memberships
  smithCert: Certs
  treasury: any
}
interface Identities {
  identities: Array<Identity>
}
interface Identity {
  index: number
  name: string
  value: IdtyValue
}
interface IdtyValue {
  data: IdtyData
  next_creatable_identity_on: number
  old_owner_key: string | null
  owner_key: string
  removable_on: number
  status: string
}
interface IdtyData {
  first_eligible_ud: number
}
interface Memberships {
  memberships: Map<number, Membership>
}
interface Membership {
  expire_on: number
}
interface Accounts {
  accounts: Map<string, Account>
  treasuryBalance: number
}
interface Account {
  random_id: string
  balance: number
  is_identity: boolean
}
interface Certs {
  applyCertPeriodAtGenesis: boolean
  certsByReceiver: Map<number, Map<number, number>>
}
interface InitialAuthorities {
  initialAuthorities: Map<number, Array<any>>
}

export default async function indexGenesis(
  app: FastifyInstance,
  genesis_info: GenesisInfo
) {
  // Read genesis json
  const filePath = resolve(cwd(), config.INDEXER_GENESIS_FILE_PATH as string)
  const resp = readFileSync(filePath)
  const data = JSON.parse(resp.toString())
  const genesis: Genesis = data.genesis.runtime
  const block0Hash = genesis_info.block0Hash
  const created_at = genesis_info.block1Date
  const created_at_str = created_at.toJSON().replace(/Z$/, '+00:00') // format creation date

  app.log.info(`Indexing Genesis`)

  // Create genesis block
  app.log.info('adding genesis block info')
  await app.sql`INSERT INTO parameters VALUES ('last_indexed_block_number', '0')`
  await app.sql`INSERT INTO parameters VALUES ('last_indexed_block_hash', ${block0Hash})`
  await app.sql`
    INSERT INTO block ${app.sql({
      number: 0,
      hash: block0Hash,
      created_at,
      data
    })}`

  // // Insert wallets (inefficient for loop)
  // app.log.info('adding wallets')
  // for (const [address, account] of Object.entries(genesis.account.accounts)) {
  //   await app.sql`
  //   INSERT INTO account ${app.sql({
  //     pubkey: address,
  //     created_at,
  //     created_on: 0
  //   })}`
  // }

  // Insert wallets (batch)
  app.log.info('adding wallets')
  var query = 'INSERT INTO account (pubkey, created_at, created_on) VALUES\n' // base of query
  // add query values
  for (const [address, account] of Object.entries(genesis.account.accounts)) {
    query += `('${address}', '${created_at_str}', 0),\n`
  }
  query = query.replace(/,\n$/, ';') // end of query
  var tmp = resolve(cwd(), './tmp/genesis_wallets.sql') // place to temporarily store query
  writeFileSync(tmp, query)
  await app.sql.file(tmp) // submit query

  // Insert identities
  app.log.info('adding identities')
  var query =
    'INSERT INTO identity (name, index, pubkey, confirmed_at, confirmed_on, validated_at, validated_on, revoked_at, revoked_on) VALUES\n'
  for (const idty of genesis.identity.identities) {
    // /!\ removable_on = 2
    // means that the identity expired but is added to the genesis to keep track of v1 identities
    if (idty.value.removable_on == 2) {
      // identity expired but can still have active certifications
      query += `('${idty.name}', '${idty.index}', '${idty.value.owner_key}', NULL, NULL, NULL, NULL, '${created_at_str}', 0),\n`
    } else {
      // identity is still valid
      query += `('${idty.name}', '${idty.index}', '${idty.value.owner_key}', '${created_at_str}', 0, '${created_at_str}', 0, NULL, NULL),\n`
    }
  }
  query = query.replace(/,\n$/, ';')
  var tmp = resolve(cwd(), './tmp/genesis_identities.sql')
  writeFileSync(tmp, query)
  await app.sql.file(tmp) // submit query

  // Insert certifications
  app.log.info('adding certifications')
  var query =
    'INSERT INTO certification (receiver_index, issuer_index, created_at, created_on) VALUES\n'
  for (const [receiver, certs_received] of Object.entries(
    genesis.cert.certsByReceiver
  )) {
    for (const [issuer, expiration_block] of Object.entries(certs_received)) {
      query += `('${receiver}', '${issuer}', '${created_at_str}', 0),\n`
    }
  }
  query = query.replace(/,\n$/, ';')
  var tmp = resolve(cwd(), './tmp/genesis_certs.sql')
  writeFileSync(tmp, query)
  await app.sql.file(tmp) // submit query

  // Insert smith
  app.log.info('adding smiths')
  for (const [smith, smith_info] of Object.entries(
    genesis.authorityMembers.initialAuthorities
  )) {
    await app.sql`
      INSERT INTO smith ${app.sql({
        idty_index: smith
      })}`
  }

  // Insert smith certifications
  app.log.info('adding smiths certifications')
  for (const [receiver, certs_received] of Object.entries(
    genesis.smithCert.certsByReceiver
  )) {
    for (const [issuer, expiration_block] of Object.entries(certs_received)) {
      await app.sql`
        INSERT INTO smith_cert ${app.sql({
          receiver_index: receiver,
          issuer_index: issuer,
          created_at,
          created_on: 0
        })}`
    }
  }
}
